"""Constants for pyacaia_async."""
from typing import Final

DEFAULT_CHAR_ID: Final = "49535343-8841-43f4-a8d4-ecbe34729bb3"
NOTIFY_CHAR_ID: Final = "49535343-1e4d-4bd9-ba61-23c647249616"
OLD_STYLE_CHAR_ID: Final = "00002a80-0000-1000-8000-00805f9b34fb"
HEADER1: Final = 0xEF
HEADER2: Final = 0xDD
HEARTBEAT_INTERVAL: Final = 5
SCALE_START_NAMES: Final = ["ACAIA", "PYXIS", "LUNAR", "PROCH"]

BATTERY_LEVEL: Final = "battery_level"
WEIGHT: Final = "weight"
UNITS: Final = "units"

GRAMS: Final = "grams"
OUNCE: Final = "ounces"



"""Client to interact with Acaia scales."""
from __future__ import annotations

import asyncio
import logging
import time

from collections.abc import Awaitable, Callable
from typing import Any

from bleak import BleakClient, BleakGATTCharacteristic, BLEDevice
from bleak.exc import BleakDeviceNotFoundError, BleakError

from .const import (
    DEFAULT_CHAR_ID,
    HEARTBEAT_INTERVAL,
    NOTIFY_CHAR_ID,
    OLD_STYLE_CHAR_ID,
)
from .exceptions import AcaiaDeviceNotFound, AcaiaError
from .const import BATTERY_LEVEL, GRAMS, WEIGHT, UNITS
from .decode import Message, Settings, decode
from .helpers import encode, encode_id, encode_notification_request

_LOGGER = logging.getLogger(__name__)


class AcaiaScale:
    """Representation of an Acaia scale."""

    _default_char_id = DEFAULT_CHAR_ID
    _notify_char_id = NOTIFY_CHAR_ID
    _msg_types = {
        "tare": encode(4, [0]),
        "startTimer": encode(13, [0, 0]),
        "stopTimer": encode(13, [0, 2]),
        "resetTimer": encode(13, [0, 1]),
        "heartbeat": encode(0, [2, 0]),
        "getSettings": encode(6, [0] * 16),
        "notificationRequest": encode_notification_request(),
    }

    def __init__(
        self,
        mac: str | None = None,
        is_new_style_scale: bool = True,
        notify_callback: Callable[[], None] | None = None,
    ) -> None:
        """Initialize the scale."""

        self._mac = mac
        self._is_new_style_scale = is_new_style_scale

        self._client: BleakClient | None = None
        self._connected = False
        self._disconnecting = False
        self._timer_running = False
        self._timestamp_last_command: float | None = None
        self._timer_start: float | None = None
        self._timer_stop: float | None = None
        self._data: dict[str, Any] = {BATTERY_LEVEL: None, UNITS: GRAMS, WEIGHT: 0.0}

        self._queue: asyncio.Queue = asyncio.Queue()

        self._msg_types["auth"] = encode_id(is_pyxis_style=is_new_style_scale)

        if not is_new_style_scale:
            # for old style scales, the default char id is the same as the notify char id
            self._default_char_id = self._notify_char_id = OLD_STYLE_CHAR_ID

        self._notify_callback: Callable[[], None] | None = notify_callback

    @property
    def mac(self) -> str:
        """Return the mac address of the scale in upper case."""
        assert self._mac
        return self._mac.upper()

    @property
    def timer_running(self) -> bool:
        """Return whether the timer is running."""
        return self._timer_running

    @timer_running.setter
    def timer_running(self, value: bool) -> None:
        """Set timer running state."""
        self._timer_running = value

    @property
    def connected(self) -> bool:
        """Return whether the scale is connected."""
        return self._connected

    @connected.setter
    def connected(self, value: bool) -> None:
        """Set connected state."""
        self._connected = value

    @property
    def data(self) -> dict[str, Any]:
        """Return the data of the scale."""
        return self._data

    @classmethod
    async def create(
        cls,
        mac: str | None = None,
        ble_device: BLEDevice | None = None,
        is_new_style_scale: bool = True,
        callback: Callable[[BleakGATTCharacteristic, bytearray], Awaitable[None] | None]
        | None = None,
    ) -> AcaiaScale:
        """Create a new scale."""
        self = cls(mac, is_new_style_scale)

        if ble_device:
            self._client = BleakClient(ble_device)
        elif mac:
            self._client = BleakClient(mac)
        else:
            raise ValueError("Either mac or bleDevice must be specified")

        if callback is None:
            callback = self.on_bluetooth_data_received
        await self.connect(callback)
        asyncio.create_task(
            self._send_heartbeats(
                interval=HEARTBEAT_INTERVAL if not is_new_style_scale else 1,
                new_style_heartbeat=is_new_style_scale,
            )
        )
        asyncio.create_task(self._process_queue())
        return self

    @property
    def msg_types(self) -> dict:
        """Return the message types."""
        return self._msg_types

    @property
    def timer(self) -> int:
        """Return the current timer value in seconds."""
        if self._timer_start is None:
            return 0
        if self._timer_running:
            return int(time.time() - self._timer_start)
        if self._timer_stop is None:
            return 0

        return int(self._timer_stop - self._timer_start)

    def new_client_from_ble_device(self, ble_device: BLEDevice) -> None:
        """Create a new client from a BLEDevice, used for Home Assistant"""
        self._client = BleakClient(ble_device)

    async def _write_msg(self, char_id: str, payload: bytearray) -> None:
        """wrapper for writing to the device."""
        try:
            if not self._connected:
                return
            assert self._client
            await self._client.write_gatt_char(char_id, payload)
            self._timestamp_last_command = time.time()
        except BleakDeviceNotFoundError as ex:
            self._connected = False
            raise AcaiaDeviceNotFound("Device not found") from ex
        except BleakError as ex:
            self._connected = False
            raise AcaiaError("Error writing to device") from ex
        except Exception as ex:
            self._connected = False
            raise AcaiaError("Unknown error writing to device") from ex

    async def _process_queue(self) -> None:
        """Task to process the queue in the background."""
        while True:
            try:
                if not self._connected:
                    while not self._queue.empty():
                        # empty the queue
                        self._queue.get_nowait()
                        self._queue.task_done()
                    return

                if self._disconnecting and self._queue.empty():
                    return

                char_id, payload = await self._queue.get()
                await self._write_msg(char_id, payload)
                self._queue.task_done()
                await asyncio.sleep(0.1)

            except asyncio.CancelledError:
                return
            except (AcaiaDeviceNotFound, AcaiaError) as ex:
                _LOGGER.debug("Error writing to device: %s", ex)
                return

    async def connect(
        self,
        callback: Callable[[BleakGATTCharacteristic, bytearray], Awaitable[None] | None]
        | None = None,
    ) -> None:
        """Initiate connection to the scale"""
        if not self._client:
            raise AcaiaError("Client not initialized")
        try:
            try:
                await self._client.connect()
            except BleakError as ex:
                _LOGGER.exception("Error during connecting to device: %s", ex)
                raise AcaiaError("Error during connecting to device") from ex

            self._connected = True
            _LOGGER.debug("Connected to Acaia Scale")

            if callback is  None:
                callback = self.on_bluetooth_data_received
            try:
                await self._client.start_notify(self._notify_char_id, callback)
                await asyncio.sleep(0.5)
            except BleakError as ex:
                _LOGGER.exception("Error subscribing to notifications: %s", ex)
                raise AcaiaError("Error subscribing to notifications") from ex

            await self.auth()

            if callback is not None:
                await self.send_weight_notification_request()

        except BleakDeviceNotFoundError as ex:
            raise AcaiaDeviceNotFound("Device not found") from ex

    async def auth(self) -> None:
        """Send auth message to scale, if subscribed to notifications returns Settings object"""
        await self._queue.put((self._default_char_id, self.msg_types["auth"]))

    async def send_weight_notification_request(self) -> None:
        """Tell the scale to send weight notifications"""

        await self._queue.put(
            (self._default_char_id, self.msg_types["notificationRequest"])
        )

    async def _send_heartbeats(
        self, interval: int = HEARTBEAT_INTERVAL, new_style_heartbeat: bool = False
    ) -> None:
        """Task to send heartbeats in the background."""
        while True:
            try:
                if not self._connected or self._disconnecting:
                    return

                _LOGGER.debug("Sending heartbeat")
                if new_style_heartbeat:
                    await self._queue.put(
                        (self._default_char_id, self.msg_types["auth"])
                    )

                await self._queue.put(
                    (self._default_char_id, self.msg_types["heartbeat"])
                )

                if new_style_heartbeat:
                    await self._queue.put(
                        (self._default_char_id, self.msg_types["getSettings"])
                    )
                await asyncio.sleep(interval)
            except asyncio.CancelledError:
                return
            except asyncio.QueueFull as ex:
                _LOGGER.debug("Error sending heartbeat: %s", ex)
                return

    async def disconnect(self) -> None:
        """Clean disconnect from the scale"""
        if not self._client:
            return
        try:
            _LOGGER.debug("Disconnecting from scale")
            self._disconnecting = True
            await self._queue.join()
            await self._client.disconnect()
            self._connected = False
            _LOGGER.debug("Disconnected from Acaia Scale")
        except BleakError as ex:
            _LOGGER.debug("Error disconnecting from device: %s", ex)

    async def tare(self) -> None:
        """Tare the scale."""
        await self._queue.put((self._default_char_id, self.msg_types["tare"]))

    async def start_stop_timer(self) -> None:
        """Start/Stop the timer."""
        if not self._timer_running:
            _LOGGER.debug('Sending "start" message.')
            await self._queue.put((self._default_char_id, self.msg_types["startTimer"]))
            self._timer_running = True
            if not self._timer_start:
                self._timer_start = time.time()
        else:
            _LOGGER.debug('Sending "stop" message.')
            await self._queue.put((self._default_char_id, self.msg_types["stopTimer"]))
            self._timer_running = False
            self._timer_stop = time.time()

    async def reset_timer(self) -> None:
        """Reset the timer."""
        await self._queue.put((self._default_char_id, self.msg_types["resetTimer"]))
        self._timer_start = None
        self._timer_stop = None

        if self._timer_running:
            await self._queue.put((self._default_char_id, self.msg_types["startTimer"]))
            self._timer_start = time.time()

    async def on_bluetooth_data_received(
        self, characteristic: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Receive data from scale."""
        msg = decode(data)[0]

        if isinstance(msg, Settings):
            self._data[BATTERY_LEVEL] = msg.battery
            self._data[UNITS] = msg.units
            _LOGGER.debug(
                "Got battery level %s, units %s", str(msg.battery), str(msg.units)
            )

        elif isinstance(msg, Message):
            self._data[WEIGHT] = msg.value
            _LOGGER.debug("Got weight %s", str(msg.value))

        if self._notify_callback is not None:
            self._notify_callback()



"""Exceptions for pyacaia_async."""
from bleak.exc import BleakDeviceNotFoundError, BleakError


class AcaiaScaleException(Exception):
    """Base class for exceptions in this module."""


class AcaiaDeviceNotFound(BleakDeviceNotFoundError):
    """Exception when no device is found."""


class AcaiaError(BleakError):
    """Exception for general bleak errors."""



"""Helper functions, taken from pyacaia."""
import logging

from bleak import BleakScanner

from .const import HEADER1, HEADER2, SCALE_START_NAMES

_LOOGER = logging.getLogger(__name__)


async def find_acaia_devices(timeout=10, scanner: BleakScanner | None = None) -> list:
    """Find ACAIA devices."""

    _LOOGER.debug("Looking for ACAIA devices")
    if scanner is None:
        async with BleakScanner() as scanner:
            return await scan(scanner, timeout)
    else:
        return await scan(scanner, timeout)


async def scan(scanner: BleakScanner, timeout) -> list:
    """Scan for devices."""
    addresses = []

    devices = await scanner.discover(timeout=timeout)
    for d in devices:
        if d.name and any(d.name.startswith(name) for name in SCALE_START_NAMES):
            print(d.name, d.address)
            addresses.append(d.address)

    return addresses


def encode(msg_type: int, payload: bytearray | list[int]) -> bytearray:
    """Encode a message to the scale."""
    byte_msg = bytearray(5 + len(payload))

    byte_msg[0] = HEADER1
    byte_msg[1] = HEADER2
    byte_msg[2] = msg_type
    cksum1 = 0
    cksum2 = 0

    for i, p_byte in enumerate(payload):
        val = p_byte & 0xFF
        byte_msg[3 + i] = val
        if i % 2 == 0:
            cksum1 += val
        else:
            cksum2 += val

    byte_msg[len(payload) + 3] = cksum1 & 0xFF
    byte_msg[len(payload) + 4] = cksum2 & 0xFF

    return byte_msg


def encode_id(is_pyxis_style=False) -> bytearray:
    """Encode the scale id."""
    if is_pyxis_style:
        payload = bytearray(
            [
                0x30,
                0x31,
                0x32,
                0x33,
                0x34,
                0x35,
                0x36,
                0x37,
                0x38,
                0x39,
                0x30,
                0x31,
                0x32,
                0x33,
                0x34,
            ]
        )
    else:
        payload = bytearray(
            [
                0x2D,
                0x2D,
                0x2D,
                0x2D,
                0x2D,
                0x2D,
                0x2D,
                0x2D,
                0x2D,
                0x2D,
                0x2D,
                0x2D,
                0x2D,
                0x2D,
                0x2D,
            ]
        )
    return encode(11, payload)


def encode_notification_request() -> bytearray:
    """Encode the notification request."""
    payload = [
        0,  # weight
        1,  # weight argument
        1,  # battery
        2,  # battery argument
        2,  # timer
        5,  # timer argument (number heartbeats between timer messages)
        3,  # key
        4,  # setting
    ]
    byte_msg = bytearray(len(payload) + 1)
    byte_msg[0] = len(payload) + 1

    for i, p_byte in enumerate(payload):
        byte_msg[i + 1] = p_byte & 0xFF

    return encode(12, byte_msg)



"""Message decoding functions, taken from pyacaia."""
import logging

from bleak import BleakGATTCharacteristic

from .const import HEADER1, HEADER2

_LOGGER = logging.getLogger(__name__)


class Message:
    """Representation of a message from the scale."""

    def __init__(self, msg_type: int, payload: bytearray | list[int]) -> None:
        self.msg_type = msg_type
        self.payload = payload
        self.value = None
        self.button = None
        self.time = None

        if self.msg_type == 5:
            self.value = self._decode_weight(payload)

        elif self.msg_type == 11:
            if payload[2] == 5:
                self.value = self._decode_weight(payload[3:])
            elif payload[2] == 7:
                self.time = self._decode_time(payload[3:])
            _LOGGER.debug(
                "heartbeat response (weight: %s, time: %s)", self.value, self.time
            )

        elif self.msg_type == 7:
            self.time = self._decode_time(payload)
            _LOGGER.debug("timer: %s", self.time)

        elif self.msg_type == 8:
            if payload[0] == 0 and payload[1] == 5:
                self.button = "tare"
                self.value = self._decode_weight(payload[2:])
                _LOGGER.debug("tare (weight: %s)", self.value)
            elif payload[0] == 8 and payload[1] == 5:
                self.button = "start"
                self.value = self._decode_weight(payload[2:])
                _LOGGER.debug("start (weight: %s)", self.value)
            elif (payload[0] == 10 and payload[1] == 7) or (
                payload[0] == 10 and payload[1] == 5
            ):
                self.button = "stop"
                self.time = self._decode_time(payload[2:])
                self.value = self._decode_weight(payload[6:])
                _LOGGER.debug("stop time: %s, weight: %s", self.time, self.value)

            elif payload[0] == 9 and payload[1] == 7:
                self.button = "reset"
                self.time = self._decode_time(payload[2:])
                self.value = self._decode_weight(payload[6:])
                _LOGGER.debug("reset time: %s, weight: %s", self.time, self.value)
            else:
                self.button = "unknownbutton"
                _LOGGER.debug("unknownbutton %s", str(payload))

        else:
            _LOGGER.debug("message: %s, payload %s", msg_type, payload)

    def _decode_weight(self, weight_payload):
        value = ((weight_payload[1] & 0xFF) << 8) + (weight_payload[0] & 0xFF)
        unit = weight_payload[4] & 0xFF
        if unit == 1:
            value /= 10.0
        elif unit == 2:
            value /= 100.0
        elif unit == 3:
            value /= 1000.0
        elif unit == 4:
            value /= 10000.0
        else:
            raise ValueError(f"unit value not in range {unit}")

        if (weight_payload[5] & 0x02) == 0x02:
            value *= -1
        return value

    def _decode_time(self, time_payload):
        value = (time_payload[0] & 0xFF) * 60
        value = value + (time_payload[1])
        value = value + (time_payload[2] / 10.0)
        return value


class Settings:
    """Representation of the settings from the scale."""

    def __init__(self, payload: bytearray) -> None:
        # payload[0] is unknown
        self.battery = payload[1] & 0x7F
        if payload[2] == 2:
            self.units = "grams"
        elif payload[2] == 5:
            self.units = "ounces"
        else:
            self.units = "grams"
        # payload[2 and 3] is unknown
        self.auto_off = payload[4] * 5
        # payload[5] is unknown
        self.beep_on = payload[6] == 1
        # payload[7-9] unknown
        _LOGGER.debug(
            "settings: battery=%s %s, auto_off=%s, beep=%s",
            self.battery,
            self.units,
            self.auto_off,
            self.beep_on,
        )
        _LOGGER.debug(
            "unknown settings: %s",
            str(
                [
                    payload[0],
                    payload[1] & 0x80,
                    payload[3],
                    payload[5],
                    payload[7],
                    payload[8],
                    payload[9],
                ]
            ),
        )


def decode(byte_msg: bytearray):
    """Return a tuple - first element is the message, or None
    if one not yet found.  Second is are the remaining
    bytes, which can be empty
    Messages are encoded as the encode() function above,
    min message length is 6 bytes
    HEADER1 (0xef)
    HEADER1 (0xdd)
    command
    length  (including this byte, excluding checksum)
    payload of length-1 bytes
    checksum byte1
    checksum byte2

    """
    msg_start = -1

    for i in range(len(byte_msg) - 1):
        if byte_msg[i] == HEADER1 and byte_msg[i + 1] == HEADER2:
            msg_start = i
            break
    if msg_start < 0 or len(byte_msg) - msg_start < 6:
        return (None, byte_msg)

    msg_end = msg_start + byte_msg[msg_start + 3] + 5

    if msg_end > len(byte_msg):
        return (None, byte_msg)

    if msg_start > 0:
        _LOGGER.debug("Ignoring %s bytes before header", i)

    cmd = byte_msg[msg_start + 2]
    if cmd == 12:
        msg_type = byte_msg[msg_start + 4]
        payload_in = byte_msg[msg_start + 5 : msg_end]
        return (Message(msg_type, payload_in), byte_msg[msg_end:])
    if cmd == 8:
        return (Settings(byte_msg[msg_start + 3 :]), byte_msg[msg_end:])

    _LOGGER.debug(
        "Non event notification message command %s %s",
        str(cmd),
        str(byte_msg[msg_start:msg_end]),
    )
    return (None, byte_msg[msg_end:])


def notification_handler(sender: BleakGATTCharacteristic, data: bytearray) -> None:
    """Sample for callback for handling incoming notifications from the scale."""
    msg = decode(data)[0]
    if isinstance(msg, Settings):
        print(f"Battery: {msg.battery}")
        print(f"Units: {msg.units}")
    elif isinstance(msg, Message):
        print(f"Weight: {msg.value}")
